// Copyright 2016, University of Colorado Boulder

/**
 * A configurable, reusable scatter plot function using d3.js.
 *
 * @author: Andrew Adare (PhET Interactive Simulations)
 */

window.scatterPlot = function() {
  'use strict';

  // Configuration variables - set or accessed via functions of the same name
  var margin = { top: 20, right: 20, bottom: 20, left: 50 };
  var width = 960 - margin.left - margin.right;
  var height = 500 - margin.top - margin.bottom;
  var axisLabel = {
    x: '',
    y: ''
  };

  var markerSize = 5;

  // Spacing (px) between axis label and axis
  var axisLabelOffset = {
    x: 40, // This is a vertical offset!
    y: 60 // This is a horizontal offset!
  };

  // Whether axis labels should be centered ('middle'), or end-aligned ('end')
  var axisLabelAlignment = {
    x: 'middle',
    y: 'middle'
  };

  // Name of key in data array for x and y dimensions
  // Should be blank by default https://github.com/phetsims/phet-io/issues/665
  var featureName = {
    x: '',
    y: ''
  };

  // Axis scale transform. Currently 'linear', 'log' (base 10), 'inverse' (1/x), or 'ln'
  var scaleTransform = {
    x: 'linear',
    y: 'linear'
  };

  // A guideline for the number of ticks on each axis. D3 may adjust slightly.
  var nTicks = {
    x: 5,
    y: 5
  };

  // Format specifier strings for x and y axes. https://github.com/d3/d3-format
  var tickFormat = {
    x: '.3',
    y: '.3'
  };

  // Accessor functions for a datapoint d
  var xValue = function( d ) { return d[ featureName.x ]; };
  var yValue = function( d ) { return d[ featureName.y ]; };

  // Optionally provide fixed axis ranges as arrays, e.g. { x: [ 0, 10 ], y: [ 0, 100 ] }
  // Note that the term "range" here actually means "domain" in the context
  // of D3 axis scales, because we're specifying the numbers to be displayed on
  // the ticks, and not the pixel coordinates of the axes.
  // Using the term "range" seems more natural for a user API than "domain".
  var fixedAxisRange = { x: null, y: null };

  // Variables below this line are for internal use and do not have direct set/get methods

  // Functions to map data value (domain) -> axis position (range)
  // TODO: call scaleTransform instead of initializing here (?)
  var xScale = d3.scaleLinear().range( [ 0, width ] );
  var yScale = d3.scaleLinear().range( [ height, 0 ] );

  // The root SVG D3 selection (svgRoot.node() is the actual DOM element)
  // @private
  var svgRoot = null;

  var xDomain = [ 0, 1 ];
  var yDomain = [ 0, 1 ];

  /**
   * Inner "instance" function object returned from the closure.
   * Updates are accomplished by binding data to a selection and then
   * using something like `selection.call(scatterPlot())`.
   * @param {Object} selection - D3 selection of DOM elements
   */
  function plot( selection ) {
    selection.each( function( data, i ) {

      // Select the svg element, if it exists. `this` is the parent element of svgRoot
      svgRoot = d3.select( this ).selectAll( 'svg' ).data( [ data ] );
      svgRoot.attr( 'class', 'plot-canvas' );

      createPlotIfNeeded();

      updateChartDimensions();
      updateAxes( data );
      updateDataPoints( data );

    } );
  }

  // If the `enter` selection is not empty (in other words, if the parent DOM
  // node doesn't already have an svg element appended), create the chart
  // frame and axes. Otherwise, this function does nothing.
  function createPlotIfNeeded() {

    // Create root node for chart
    var gEnter = svgRoot.enter().append( 'svg' )
      .attr( 'width', width )
      .attr( 'height', height )
      .append( 'g' );

    // Offset chart element by margins
    gEnter
      .attr( 'transform', 'translate(' + margin.left + ',' + margin.top + ')' );

    // Create plot axes
    gEnter.append( 'g' )
      .call( d3.axisBottom( xScale ) )
      .attr( 'class', 'x axis' );
    gEnter.append( 'g' )
      .call( d3.axisLeft( yScale ) )
      .attr( 'class', 'y axis' );

    var xAxisLength = width - margin.left - margin.right;
    var yAxisLength = height - margin.bottom - margin.top;

    // Provide initial axis configuration so the axes can be rendered without data
    gEnter.select( '.x.axis' )
      .attr( 'transform', 'translate(' + 0 + ',' + yAxisLength + ')' )
      .call( d3.axisBottom( d3.scaleLinear()
        .domain( xDomain )
        .range( [ 0, xAxisLength ] ) )
        .ticks( nTicks.x, tickFormat.x ) );
    gEnter.select( '.y.axis' )
      .attr( 'transform', 'translate(' + 0 + ',' + 0 + ')' )
      .call( d3.axisLeft( d3.scaleLinear()
        .domain( yDomain )
        .range( [ yAxisLength, 0 ] ) )
        .ticks( nTicks.y, tickFormat.y ) );

    // x-axis label
    gEnter.select( '.x.axis' ).append( 'text' )
      .attr( 'class', 'x axis label' )
      .attr( 'x', axisLabelAlignment.x === 'end' ? xAxisLength : xAxisLength / 2 )
      .attr( 'y', axisLabelOffset.x )
      .style( 'text-anchor', axisLabelAlignment.x )
      .text( axisLabel.x );

    // y-axis label
    gEnter.select( '.y.axis' ).append( 'text' )
      .attr( 'class', 'y axis label' )
      .attr( 'transform', 'rotate(-90)' )
      .attr( 'x', axisLabelAlignment.y === 'end' ? 0 : -yAxisLength / 2 ) // Vertical, since after rotation
      .attr( 'y', -axisLabelOffset.y ) // Horizontal, since after rotation
      .style( 'text-anchor', axisLabelAlignment.y )
      .text( axisLabel.y );
  }

  function updateChartDimensions() {
    // Update the outer dimensions
    svgRoot
      .attr( 'width', width )
      .attr( 'height', height );

    // Update the inner dimensions
    svgRoot.select( 'g' )
      .attr( 'transform', 'translate(' + margin.left + ',' + margin.top + ')' );
  }

  // Update data scale mapping functions and axis objects
  function updateAxes( data ) {

    // Extend axis range outside interval spanned by data: padFactor*(max - min)
    var padFactor = 0;

    if ( data && data.length > 0 ) {
      if ( !fixedAxisRange.x ) {
        xDomain = displayInterval( data, xValue, padFactor );
      }
      if ( !fixedAxisRange.y ) {
        yDomain = displayInterval( data, yValue, padFactor );
      }
    }

    // Only 'linear' axis scales are allowed if domain includes zero
    if ( xDomain[ 0 ] <= 0 && xDomain[ 1 ] >= 0 && scaleTransform.x !== 'linear' ) {
      scaleTransform.x = 'linear';
      xScale = d3.scaleLinear();
    }
    if ( yDomain[ 0 ] <= 0 && yDomain[ 1 ] >= 0 && scaleTransform.y !== 'linear' ) {
      scaleTransform.y = 'linear';
      yScale = d3.scaleLinear();
    }

    xScale
      .domain( xDomain )
      .nice()
      .range( [ 0, width - margin.left - margin.right ] );
    yScale
      .domain( yDomain )
      .nice()
      .range( [ height - margin.top - margin.bottom, 0 ] );

    var g = svgRoot.select( 'g' );
    g.select( '.x.axis' )
      .attr( 'transform', 'translate(' + 0 + ',' + yScale.range()[ 0 ] + ')' )
      .call(
        d3.axisBottom( xScale )
          .ticks( nTicks.x, tickFormat.x )
      );
    g.select( '.y.axis' )
      .attr( 'transform', 'translate(' + 0 + ',' + 0 + ')' )
      .call(
        d3.axisLeft( yScale )
          .ticks( nTicks.y, tickFormat.y )
      );

    // Update axis labels
    g.select( '.x.axis.label' )
      .text( axisLabel.x );

    g.select( '.y.axis.label' )
      .text( axisLabel.y );
  }

  // Update data points: manage the three selections (enter, update, exit)
  function updateDataPoints( data ) {
    var cx = function( d ) { return margin.left + xScale( d[ featureName.x ] ); };
    var cy = function( d ) { return margin.top + yScale( d[ featureName.y ] ); };

    if ( !data ) { return; }


    var visibleData = data.filter( function( d ) {
      return ( d.visible );
    } );

    // Remove invalid points before updating graph
    var validData = visibleData.filter( function( d ) {
      var dx = d[ featureName.x ];
      var dy = d[ featureName.y ];
      return !( dx === null || typeof dx === 'undefined' || dy === null || typeof dy === 'undefined' );
    } );


    if ( validData.length !== visibleData.length && featureName.x !== 'x' && featureName.y !== 'y' ) {
      window.toast( 'Some data points are not shown because they are empty.' );
    }

    var points = svgRoot.selectAll( '.datapoint' ).data( validData ); // update
    points.exit().remove(); // exit
    points.enter().append( 'circle' ) // enter
      .attr( 'class', 'datapoint' )
      .attr( 'r', markerSize )
      .merge( points )
      .attr( 'cx', cx )
      .attr( 'cy', cy )
      .attr( 'opacity', function( d ) {
        if ( typeof d.visible === 'undefined' ) { return 1; }
        return d.visible ? 1 : 0;
      } );
  }

  /**
   * Return a 2-element array with the bounds of the axis interval to display
   *
   * @param  {Array.<Object>} data - Array of datapoints
   * @param  {function} accessor - function used by d3 to extract the x or y value from a datapoint object
   * @param  {number} padFactor - Extend interval by padFactor*(max - min) on either end
   *
   * @returns {Array.<number>} Axis display bounds
   */
  function displayInterval( data, accessor, padFactor ) {

    // Include only visible subset of data in display interval
    var visibleData = [];
    data.forEach( function( d ) {
      if ( d.visible || typeof d.visible === 'undefined' ) {
        visibleData.push( d );
      }
    } );

    var extent = d3.extent( visibleData, accessor );
    var min = extent[ 0 ];
    var max = extent[ 1 ];
    var pad = padFactor * ( max - min );

    // Handle case where min and max are the same
    if ( min === max ) {
      extent = min ? [ 0.5 * min, 1.5 * min ] : [ 0, 1 ];
    }
    else {
      extent = [ min - pad, max + pad ];
    }
    return extent;
  }

  // Chainable getter/setter methods (each is both)
  // This is an idiomatic D3 pattern (but not typical for PhET). All built-in
  // D3 components work in this way.

  /**
   * Set/get axis label text
   *
   * @param  {[Object]} label - object like {x: 'oranges', y: 'apples'} or just { y: 'apples' }
   *
   * @returns {Object} Current labels if called without args
   */
  plot.axisLabel = function( label ) {
    if ( !arguments.length ) { return axisLabel; }

    if ( label.x ) {
      axisLabel.x = label.x;
    }
    if ( label.y ) {
      axisLabel.y = label.y;
    }
    return plot;
  };

  /**
   * Set/get spacing between axis labels and axes in px
   *
   * @param  {[Object]} offset - object like {x: 10, y: 20} or just { x: 10 }
   *
   * @returns {Object} Current labels if called without args
   */
  plot.axisLabelOffset = function( offset ) {
    if ( !arguments.length ) { return axisLabelOffset; }

    if ( offset.x ) {
      axisLabelOffset.x = offset.x;
    }
    if ( offset.y ) {
      axisLabelOffset.y = offset.y;
    }
    return plot;
  };

  /**
   * Set/get alignment of axis labels - choose 'middle' for centered, or 'end' for right-justified
   *
   * @param  {[Object]} alignment - object like {x: 'end', y: 'end'} or just { x: 'middle' }
   *
   * @returns {Object} Current labels if called without args
   */
  plot.axisLabelAlignment = function( alignment ) {
    if ( !arguments.length ) { return axisLabelAlignment; }

    if ( alignment.x === 'middle' || alignment.x === 'end' ) {
      axisLabelAlignment.x = alignment.x;
    }
    if ( alignment.y === 'middle' || alignment.y === 'end' ) {
      axisLabelAlignment.y = alignment.y;
    }
    return plot;
  };

  /**
   * Set (or get) fixed ranges for axis display
   * @param {Object} range - object like {x: [xmin,xmax], y: [ymin,ymax]}
   */
  plot.fixedAxisRange = function( range ) {
    if ( !arguments.length ) { return fixedAxisRange; }
    if ( range.x && range.x.length === 2 ) {
      fixedAxisRange.x = range.x;
      xDomain = range.x;
    }
    if ( range.y && range.y.length === 2 ) {
      fixedAxisRange.y = range.y;
      yDomain = range.y;
    }
    return plot;
  };

  plot.height = function( _ ) {
    if ( !arguments.length ) { return height; }
    height = _;
    return plot;
  };

  // TODO make margins individually settable
  plot.margin = function( _ ) {
    if ( !arguments.length ) { return margin; }
    margin = _;
    return plot;
  };

  plot.markerSize = function( _ ) {
    if ( !arguments.length ) { return markerSize; }
    markerSize = _;
    return plot;
  };

  plot.width = function( _ ) {
    if ( !arguments.length ) { return width; }
    width = _;
    return plot;
  };

  plot.x = function( _ ) {
    if ( !arguments.length ) { return xValue; }
    xValue = _;
    return plot;
  };

  plot.y = function( _ ) {
    if ( !arguments.length ) { return yValue; }
    yValue = _;
    return plot;
  };

  /**
   * Set/get name of feature(s) to be plotted. The name should correspond to a
   * key in every datapoint, but validation is done elsewhere.
   *
   * @param  {[Object]} name - object like {x: 'apples', y: 'oranges'} or just { x: 'apples' }
   *
   * @returns {Object} Current feature keys if called without args
   */
  plot.featureName = function( name ) {
    if ( !arguments.length ) { return featureName; }

    if ( name.x ) {
      featureName.x = name.x;
    }
    if ( name.y ) {
      featureName.y = name.y;
    }

    return plot;
  };

  plot.nTicks = function( ticks ) {
    if ( !arguments.length ) { return nTicks; }

    if ( ticks.x ) {
      nTicks.x = ticks.x;
    }
    if ( ticks.y ) {
      nTicks.y = ticks.y;
    }

    return plot;
  };

  plot.tickFormat = function( format ) {
    if ( !arguments.length ) { return tickFormat; }

    if ( format.x ) {
      tickFormat.x = format.x;
    }
    if ( format.y ) {
      tickFormat.y = format.y;
    }

    return plot;
  };

  // @param [Object] transform - a hash like {x: 'linear', y: 'log'} (entries not both required)
  plot.scaleTransform = function( transform ) {
    if ( !arguments.length ) { return scaleTransform; }

    // TODO refactor
    if ( transform.x === 'log' ) { // TODO: log --> log10?
      xScale = d3.scaleLog().base( 10 );
    }
    else if ( transform.x === 'inverse' ) {
      xScale = d3.scalePow().exponent( -1 );
    }
    else if ( transform.x === 'linear' ) {
      xScale = d3.scaleLinear();
    }
    else if ( transform.x === 'ln' ) {
      xScale = d3.scaleLlog().base( Math.E );
    }
    if ( transform.y === 'log' ) {
      yScale = d3.scaleLog().base( 10 );
    }
    else if ( transform.y === 'inverse' ) {
      yScale = d3.scalePow().exponent( -1 );
    }
    else if ( transform.y === 'linear' ) {
      yScale = d3.scaleLinear();
    }
    else if ( transform.y === 'ln' ) {
      yScale = d3.scaleLlog().base( Math.E );
    }

    // TODO validate
    if ( transform.x ) {
      scaleTransform.x = transform.x;
    }
    if ( transform.y ) {
      scaleTransform.y = transform.y;
    }

    // Make the y axis increase vertically, opposite to the SVG convention
    xScale = xScale.range( [ 0, width ] );
    yScale = yScale.range( [ height, 0 ] );

    return plot;
  };

  return plot;
};