// Copyright 2016, University of Colorado Boulder

/**
 * Qualitative bar chart with one bar to serve as a voltage meter.
 *
 * @param  {string} divId - ID of div to place this graphic
 * @param  {number} voltage - a number in -1 to +1
 * @param  {Object} options - configuration options
 *
 */
window.voltageBar = function( divId, voltage, options ) {
  'use strict';

  if ( !options ) {
    options = {};
  }

  var selection = d3.select( divId );

  if ( selection.empty() ) {
    console.error( 'D3 selection for ' + divId + ' is empty' );
    return;
  }

  // Defaults
  var boundingRect = selection.node().getBoundingClientRect();
  var w = boundingRect.width;
  var h = boundingRect.height;
  var margin = {
    top: 10,
    right: 5,
    bottom: 5,
    left: 25
  };

  // Options override defaults
  if ( options.width ) {
    w = options.width;
  }
  if ( options.height ) {
    h = options.height;
  }
  if ( options.margin ) {
    margin = options.margin;
  }

  var width = w - margin.left - margin.right;
  var height = h - margin.top - margin.bottom;

  var yScale = d3.scaleLinear()
    .domain( [ -1, 1 ] )
    .range( [ height, 0 ] );

  var yAxis = d3.axisLeft()
    .scale( yScale )
    .tickValues( [ -1, 0, +1 ] )
    .tickFormat( function( y ) {
      return y < 0 ? '-' : y > 0 ? '+' : '0'; // Map a string to the tick values
    } );

  var svg = d3.select( divId ).append( 'svg' )
    .attr( 'width', width + margin.left + margin.right )
    .attr( 'height', height + margin.top + margin.bottom );

  var bar = svg.selectAll( 'g' )
    .data( [ voltage ] )
    .enter().append( 'g' )
    .attr( 'transform', 'translate(' + margin.left + ',' + margin.top + ')' );

  // This block is a bit confusing because SVG rectangles must be drawn top-down
  // (no negative heights) so the y-origin depends on whether voltage is positive
  // or negative. The inverted-y SVG convention is handled in yScale.
  bar.append( 'rect' )
    .attr( 'class', 'voltage-bar' )
    .attr( 'width', 10 )
    .attr( 'y', function( v ) {
      return yScale( Math.max( 0, v ) );
    } )
    .attr( 'height', function( v ) {
      return Math.abs( yScale( v ) - yScale( 0 ) );
    } );

  svg.append( 'g' )
    .attr( 'class', 'voltage-bar-axis' )
    .attr( 'transform', 'translate(' + margin.left + ',' + margin.top + ')' )
    .call( yAxis );
};
