// Copyright 2015-2017, University of Colorado Boulder

/**
 * Instance proxies severs as the main wrapper to visualize PhET-iO customization and manipulation.
 * This file is the responsible for communicating with the sim and updating the panel with all tandem instances from the
 * sim. These tandems are made into interactive divs in `createInstanceProxyDiv` and then are added to the right panel as
 * ui components that exemplify each component's interaction.
 */
(function() {
  'use strict';

  // Batch DOM ops so forces-and-motion-basics doesn't take 30 minutes to launch
  var toAppend = [];
  var toRemove = [];

  var main = function() {

    // Keep the control panel at the top of the screen but scrollable left and right
    $( window ).scroll( function( event ) {
      $( '#controls' ).css( 'margin-left', -1 - $( document ).scrollLeft() );
    } );

    var main = function() {

      // Initialize the screens
      var sim = WrapperUtils.getSim( 'john-travoltage', '1.4.0-phetiodev.1' );

      var panel = document.getElementById( 'panel' );

      // Organizational checkboxes above the tandems
      var showCustomizableCheckBox = document.getElementById( 'showCustomizableCheckBox' );
      var showNodesCheckbox = document.getElementById( 'showNodesCheckbox' );
      var showAllCheckBox = document.getElementById( 'showAllCheckBox' );
      var showPropertiesCheckBox = document.getElementById( 'showPropertiesCheckBox' );

      var initCheckBoxValues = function() {
        showAllCheckBox.checked = true;
        showNodesCheckbox.checked = false;
        showCustomizableCheckBox.checked = false;
        showPropertiesCheckBox.checked = false;
      };
      document.addEventListener( 'DOMContentLoaded', initCheckBoxValues );
      initCheckBoxValues();

      /**
       * Given a string, and a type function, return whether the type is of type typeName (direct or a child of the type)
       * Uses indexOf logic to handle parametric types like TProperty that look like 'TProperty.<TBoolean>'
       * @param {string} typeName
       * @param {object} type - the TType API from phetio.getAPIForType
       * @returns {boolean}
       */
      var isType = function( typeName, type ) {
        assert && assert( typeName[ 0 ] === 'T', 'Wrapper Type names start with an uppercase T' );
        if ( type.typeName.indexOf( typeName ) === 0 ) {
          return true;
        }
        else {
          if ( type.supertype ) {
            return isType( typeName, type.supertype );
          }
          else {
            return false;
          }
        }
      };

      /**
       * Update instances synchronously, this prevents multiple removal issues,
       * see https://github.com/phetsims/phet-io/issues/781
       */
      var updateInstances = function() {
        var delta = toAppend.length + toRemove.length;
        for ( var i = 0; i < toAppend.length; i++ ) {
          panel.appendChild( toAppend[ i ].div );
          panel.appendChild( toAppend[ i ].hr );
        }
        toAppend.length = 0;

        for ( i = 0; i < toRemove.length; i++ ) {

          // If the Element Node is in the panel, remove it
          if ( panel.contains( toRemove[ i ].div ) ) {
            panel.removeChild( toRemove[ i ].div );
            panel.removeChild( toRemove[ i ].hr );
          }

          // Always remove it from the master instances object
          delete instances[ toRemove[ i ].phetioID ];
        }
        toRemove.length = 0;

        if ( delta > 0 ) {
          updateControlPanel();
        }
      };

      /**
       * Add this one instance to the ui if applicable.
       * @param {string} phetioID - the tandem id
       */
      var filterAndAppend = function( phetioID ) {

        // Allow prefixes to filter out what is shown
        var prefixes = QueryStringMachine.get( 'prefixes', {
          type: 'array',
          elementSchema: {
            type: 'string'
          },
          defaultValue: []
        } );

        var value = instances[ phetioID ];

        var isPropertyOrDerivedProperty = isType( 'TProperty', value.type ) || isType( 'TDerivedProperty', value.type );

        // Show something if it passes any rule from the checkboxes
        var show = (showAllCheckBox.checked) ||
                   (showNodesCheckbox.checked && isType( 'TNode', value.type )) ||
                   (showCustomizableCheckBox.checked && value.customizable) ||
                   (showPropertiesCheckBox.checked && value.type && isPropertyOrDerivedProperty);

        if ( prefixes.length > 0 ) {
          var matchedPrefix = false;
          for ( var i = 0; i < prefixes.length; i++ ) {
            var prefix = prefixes[ i ];
            if ( phetioID.indexOf( prefix ) === 0 ) {
              matchedPrefix = true;
            }
          }

          if ( !matchedPrefix ) {
            show = false;
          }
        }

        if ( show ) {

          value.hr = document.createElement( 'hr' );

          // Control for one item
          toAppend.push( value );
          updateInstances();
        }
      };

      /**
       * Remove the whole panel and redraw it with updated instances
       */
      var filter = function() {
        while ( panel.firstChild ) {
          panel.removeChild( panel.firstChild );
        }
        var keys = _.keys( instances );
        keys.forEach( function( key ) {
          filterAndAppend( key );
        } );
        updateControlPanel();
      };

      showAllCheckBox.addEventListener( 'change', function() {
        if ( showAllCheckBox.checked ) {
          showCustomizableCheckBox.checked = false;
          showNodesCheckbox.checked = false;
          showPropertiesCheckBox.checked = false;
        }
        filter();
      } );
      showCustomizableCheckBox.addEventListener( 'change', function() {
        if ( showCustomizableCheckBox.checked ) {
          showAllCheckBox.checked = false;
        }
        filter();
      } );
      showNodesCheckbox.addEventListener( 'change', function() {
        if ( showNodesCheckbox.checked ) {
          showAllCheckBox.checked = false;
        }
        filter();
      } );
      showPropertiesCheckBox.addEventListener( 'change', function() {
        if ( showPropertiesCheckBox.checked ) {
          showAllCheckBox.checked = false;
        }
        filter();
      } );

      window.simIFrameClient = new SimIFrameClient( document.getElementById( 'source' ) );

      // Map phetioID => type
      var instances = {};

      /**
       * updates the totals in the control panel based on the number and type of instances in the right panel.
       */
      var updateControlPanel = function() {
        var count = Object.keys( instances ).length;
        document.getElementById( 'showAllCheckBoxLabel' ).innerHTML = 'All (' + count + ')';

        var nodeCount = _.keys( instances ).filter( function( key ) { return isType( 'TNode', instances[ key ].type ); } ).length;
        document.getElementById( 'showNodesCheckBoxLabel' ).innerHTML = 'Nodes (' + nodeCount + ')';

        var customizableCount = _.keys( instances ).filter( function( key ) { return instances[ key ].customizable; } ).length;
        document.getElementById( 'showCustomizableCheckBoxLabel' ).innerHTML = 'Customizable (' + customizableCount + ')';

        var propertyCount = _.keys( instances ).filter( function( key ) {
          var value = instances[ key ];
          assert && assert( !!value.type, 'value should have a type' );

          return isType( 'TProperty', value.type ) || isType( 'TDerivedProperty', value.type );
        } ).length;
        document.getElementById( 'showPropertiesCheckBoxLabel' ).innerHTML = 'Properties (' + propertyCount + ')';
      };

      /**
       * On instance added function that is passed into the phet-io framework. It is called each time an instance is
       * created/added to the sim and ready for phet-io customization.
       * @param {String} phetioID
       * @param {object} type - the TType api, see phetio.getAPIForType()
       * @param {Object} stateObject - A snapshot of the instance
       * @param {Object} options
       */
      var instanceAdded = function( phetioID, type, stateObject, options ) {
        assert && assert( type, 'type should be defined' );
        instances[ phetioID ] = window.createInstanceProxyDiv( phetioID, type, isType, window.simIFrameClient, stateObject, options );
        instances[ phetioID ].phetioID = phetioID;
        updateControlPanel();

        filterAndAppend( phetioID );
      };

      /**
       * This is the main launch point for the wrapper. This call sets up all of the listeners associated with phet-io
       * and then launches the simulation.
       */
      window.simIFrameClient.launchSim( sim.URL, {
        onPhETiOInitialized: function() {

          // When phet-io is initialized, but before this sim is initialized, add a listener for instances being added
          // and removed. Do this synchronously.
          window.simIFrameClient.invoke( 'phetio', 'addInstanceAddedListener', [ instanceAdded ], function() {
            window.simIFrameClient.invoke( 'phetio', 'addInstanceRemovedListener', [ function( phetioID ) {
              assert && assert( instances[ phetioID ] !== undefined, 'instance should be defined: ' + phetioID );
              toRemove.push( instances[ phetioID ] );
              updateInstances();
            } ] );
          } );
        }
      } );

      var additionalQueryParameters = '';
      if ( window.location.search.length > 0 ) {
        additionalQueryParameters = '&' + window.location.search.substring( 1 );
      }


      /**
       * Gather all non-default settings, combine with in-sim customizations, and then save as HTML
       * @param {boolean} appendToTextBox
       * @param {boolean} launch
       */
      var generateHTMLAndMaybeLaunch = function( appendToTextBox, launch ) {

        // Send a command to the sim that requests the velocity value in the model.
        window.simIFrameClient.invoke( 'phetio', 'getState', [], function( state ) {

          // { phetioID:string, method:string, [args:objects:optional] }
          var expressions = [ {
            phetioID: 'phetio',
            method: 'setState',
            args: [ state ]
          } ];

          // Start simple, let's save node visibility
          _.keys( instances ).forEach( function( id ) {
            var value = instances[ id ];

            // TODO: When the checkbox values get sync'ed then we can send state for every one, not just deselected.
            if ( isType( 'TNode', value.type ) && !value.visibleCheckBox.checked ) {
              expressions.push( {
                phetioID: id,
                method: 'setVisible',
                args: [ value.visibleCheckBox.checked ]
              } );
            }
            if ( isType( 'TImage', value.type ) && value.base64 ) {
              expressions.push( {
                phetioID: id,
                method: 'setImage',
                args: [ value.base64 ]
              } );
            }

            if ( isType( 'TNode', value.type ) && value.detachButton.pressed ) {
              expressions.push( {
                phetioID: id,
                method: 'detach',
                args: []
              } );
            }

            // Send pickable values.  // TODO: make sure pickable checkboxes updated from sim or these will be wrong.
            if ( isType( 'TNode', value.type ) ) {
              var pickableValue = value.pickableFalseCheckBox.checked ? false :
                                  value.pickableNullCheckBox.checked ? null :
                                  value.pickableTrueCheckBox ? true :
                                  assert && assert( false ) && 'Error';
              console.log( [ id, 'setPickable', [ pickableValue ] ] );
              expressions.push( {
                phetioID: id,
                method: 'setPickable',
                args: [ pickableValue ]
              } );
            }

            if ( value && value.type &&
                 [ 'TText', 'TMultiLineText', 'TRichText' ].indexOf( value.type.typeName ) >= 0 &&
                 value.textInput.value ) {
              expressions.push( {
                phetioID: id,
                method: 'setText',
                args: [ value.textInput.value ]
              } );
            }
          } );
          console.log( JSON.stringify( expressions, null, 2 ) );

          var replaceAll = function( string, search, replacement ) {
            return string.split( search ).join( replacement );
          };

          /**
           * Fills out the template html phet-io sim with the data from the current sim in instance proxies.
           * @param templateText
           */
          var useTemplate = function( templateText ) {

            // The source that the iframe will have.
            var iframesrc = sim.URL;

            // The extra scripts needed to be added to the template, different for built/requirejs runs
            var extraScriptsString = '';

            // spot or phet-io.colorado.edu
            if ( WrapperUtils.isBuilt ) {

              var pathToLib = '';

              // Production deployed, on phet-server the path begins with /sims/
              if ( window.location.pathname.indexOf( '/sims/' ) === 0 ) {
                pathToLib = WrapperUtils.PRODUCTION_SIM_PATH + 'john-travoltage/1.4.0-phetiodev.1/';
              }

              // Spot deploy, for RC or Dev versions
              else if ( window.location.pathname.indexOf( '/physics/phet/dev' ) === 0 ) {

                var rootOfSimVersion = WrapperUtils.SPOT_SIM_PATH + 'john-travoltage/1.4.0-phetiodev.1/';

                // Get the name of the runnable
                var urlParts = sim.URL.split( '/' );
                var runnableName = urlParts[ urlParts.length - 1 ];

                // This has to be different from WrapperUtils for spot, because we need an absolute path.
                iframesrc = rootOfSimVersion + runnableName;

                pathToLib = rootOfSimVersion;
              }

              // For unpublished, local builds
              else if ( window.location.pathname.indexOf( '/build/' ) > 0 ) {
                pathToLib = '../../';
              }

              // Not exactly sure what the case is, perhaps a 3rd party is running it, just give the sim URL
              else {
                iframesrc = sim.URL;
              }

              // Add the import for phet-io.js
              extraScriptsString = '<script type="text/javascript" src="' + pathToLib + 'lib/phet-io.js' + '"></script>';
            }

            // requirejs mode
            else {

              // Add the needed requirejs scripts that are normally in phet-io.js
              // Note: these should be the same as the lib files in chipper/copySupplementalPhETIOFiles.js
              extraScriptsString += '<script type="text/javascript" src="../common/js/WrapperTypes.js"></script>\n';
              extraScriptsString += '<script type="text/javascript" src="../common/js/assert.js"></script>\n';
              extraScriptsString += '<script type="text/javascript" src="../common/js/SimIFrameClient.js"></script>\n';
              extraScriptsString += '<script type="text/javascript" src="../common/js/WrapperUtils.js"></script>\n';
              extraScriptsString += '<script type="text/javascript" src="../../query-string-machine/js/QueryStringMachine.js"></script>\n';
            }

            iframesrc += additionalQueryParameters;

            templateText = replaceAll( templateText, '{{EXPRESSIONS}}', encodeURIComponent( JSON.stringify( expressions, null, 2 ) ) );
            templateText = replaceAll( templateText, '{{IFRAMESRC}}', iframesrc );
            templateText = replaceAll( templateText, '{{EXTRA_PHET_SCRIPTS}}', extraScriptsString );

            // When pressing launch, we do not want to pass through query parameters, or they will be duplicated and
            // cause validation errors in Query String Machine. When pressing generate html, we do want to pass through
            // query parameters through, so that most generated wrappers will support passed through query parameters.
            // We used a new pattern instead of {{}} so that the HTML file can remain as parseable HTML/JS and not
            // become broken in a reformat.
            templateText = replaceAll( templateText, 'passThroughQueryParameters: true // {{DO NOT EDIT: THIS ENTIRE LINE IS REPLACED IN THE TEMPLATE INSTANTIATION}}',
              launch ? 'passThroughQueryParameters: false' :
              'passThroughQueryParameters: true'
            );

            if ( appendToTextBox ) {
              document.getElementById( 'textBox' ).value = templateText;
            }

            // Pop open the new document for testing.
            if ( launch ) {
              var popup = window.open();
              if ( !popup ) {
                window.alert( 'Popup was blocked, please allow popups to launch the customized simulation.' );
              }
              else {
                popup.document.write( templateText );
              }
            }
          };

          readFile( 'instance-proxies-launch-template.html', useTemplate );
        } );
      };

      // Assign listeners to the appropriate buttons to export the state of instance proxies to another sim.
      document.getElementById( 'generateHTMLButton' ).addEventListener( 'click', function() {
        generateHTMLAndMaybeLaunch( true, false );
      } );
      document.getElementById( 'launchButton' ).addEventListener( 'click', function() {
        generateHTMLAndMaybeLaunch( false, true );
      } );
    };

    main();
  };

  // Wait until jquery loaded on IE11
  if ( typeof $ === 'undefined' ) {
    var id = setTimeout( function() {
      if ( typeof $ === 'function' ) {
        clearTimeout( id );
        main();
      }
    }, 30 );
  }
  else {
    main();
  }
})();